/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Marco Maccaferri - initial implementation
 */

#include <propeller.h>

#include "vga_driver.h"
#include "graphics_renderer.h"
#include "nes_gamepad_ctrl.h"

extern uint32_t palette_def[];
extern uint32_t tiles_def[];
extern uint32_t sprites_def[];
extern uint16_t map_def[];

int bg_ptr, bg_col_left, bg_col_right, bg_x;

void initialize_background() {
    int i, ptr;

    for (i = 0; i < 32 * 34; i++)
        tile_map[i].tile = 16;

    bg_ptr = 0;

    for (bg_col_right = 0; bg_col_right < 34; bg_col_right++) {
        ptr = 23 * 34 + bg_col_right;
        for (i = 0; i < 10; i++) {
            tile_map[ptr].value = map_def[bg_ptr++];
            ptr -= 34;
        }
    }

    bg_col_left = 0;
    bg_col_right = 33;

    bg_x = 8;

    offscreen_param.x = bg_x;
}

void scroll_background_left() {
    int i, ptr;

    bg_x += 4;
    if (bg_x == 16) {
        ptr = 23 * 34;
        for (i = 0; i < 10; i++) {
            memmove(&tile_map[ptr], &tile_map[ptr + 1], 33 * 2);
            ptr -= 34;
        }

        if (++bg_col_right >= 50) {
            bg_col_right = 0;
        }

        if (++bg_col_left >= 50) {
            bg_col_left = 0;
        }

        ptr = 23 * 34 + 33;
        bg_ptr = bg_col_right * 10;

        for (i = 0; i < 10; i++) {
            tile_map[ptr].value = map_def[bg_ptr++];
            ptr -= 34;
        }

        bg_x = 8;
    }

    offscreen_param.x = bg_x;
}

void scroll_background_right() {
    int i, ptr;

    bg_x -= 4;
    if (bg_x == 0) {
        ptr = 23 * 34;
        for (i = 0; i < 10; i++) {
            memmove(&tile_map[ptr + 1], &tile_map[ptr], 33 * 2);
            ptr -= 34;
        }

        if (--bg_col_left < 0) {
            bg_col_left = 49;
        }

        if (--bg_col_right < 0) {
            bg_col_right = 49;
        }

        ptr = 23 * 34;
        bg_ptr = bg_col_left * 10;

        for (i = 0; i < 10; i++) {
            tile_map[ptr].value = map_def[bg_ptr++];
            ptr -= 34;
        }

        bg_x = 8;
    }

    offscreen_param.x = bg_x;
}

#define WALK_DELAY    5

#define HARRY_PALETTE 2

#define HARRY_FRONT_A 1
#define HARRY_PROFILE_A 2
#define HARRY_WALK1_A 3
#define HARRY_WALK2_A 4
#define HARRY_WALK3_A 5
#define HARRY_WALK4_A 6
#define HARRY_FRONT_B 7
#define HARRY_PROFILE_B 8
#define HARRY_WALK1_B 9
#define HARRY_WALK2_B 10
#define HARRY_WALK3_B 11
#define HARRY_WALK4_B 12

uint32_t    play_counter;
uint32_t    play_state;

uint16_t    harry_x;
uint16_t    harry_y;
uint8_t     harry_walk_delay;
uint8_t     harry_walk_state;
uint8_t     harry_walk_direction;

void animate_harry() {
    if (harry_walk_state != 0) {
        if (harry_walk_direction == 0) {
            if (harry_x < 256 - 64)
                harry_x += 4;
            else
                scroll_background_left();
        }
        else {
            if (harry_x > 16 + 64)
                harry_x -= 4;
            else
                scroll_background_right();
        }
    }


    switch(harry_walk_state) {
        // stop
        case 0:
            sprite_show(0, HARRY_PROFILE_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_PROFILE_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = 0;
            break;

        // begin walk loop states
        case 1:
            sprite_show(0, HARRY_WALK1_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK1_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 2:
            sprite_show(0, HARRY_WALK2_A, HARRY_PALETTE, harry_x, harry_y-1, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK2_B, HARRY_PALETTE, harry_x, harry_y-1 + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 3:
            sprite_show(0, HARRY_WALK1_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK1_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 4:
            sprite_show(0, HARRY_PROFILE_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_PROFILE_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 5:
            sprite_show(0, HARRY_WALK3_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK3_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 6:
            sprite_show(0, HARRY_WALK4_A, HARRY_PALETTE, harry_x, harry_y-1, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK4_B, HARRY_PALETTE, harry_x, harry_y-1 + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 7:
            sprite_show(0, HARRY_WALK3_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_WALK3_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state++;
            break;

        case 8:
            sprite_show(0, HARRY_PROFILE_A, HARRY_PALETTE, harry_x, harry_y, 0, harry_walk_direction);
            sprite_show(1, HARRY_PROFILE_B, HARRY_PALETTE, harry_x, harry_y + 16, 0, harry_walk_direction);
            harry_walk_delay = WALK_DELAY;
            harry_walk_state = 1;
            break;
        // end walk loop states
    }
}

void print_at(int x, int y, int palette, const char * s) {
    int c, ptr = y * 34 + x;

    while(*s) {
        c = *s++;
        if (c >= '0' && c <= '0')
            c = (c - 'A') + 17;
        else if (c >= 'A' && c <= 'Z')
            c = (c - 'A') + 27;
        else
            c = 0;
        tile_map[ptr].palette = palette;
        tile_map[ptr++].tile = c;
    }
}

int main() {
    start_vga_driver();
    start_renderers(palette_def, tiles_def, sprites_def);

    offscreen_param.y_lock = 104;

    play_counter = 0;
    play_state = 0;

    initialize_background();
    print_at(3, 4, 3, "GAME ENGINE ANIMATION DEMO");

    harry_x = 136;
    harry_y = 166;
    harry_walk_delay = 0;
    harry_walk_state = 0;
    harry_walk_direction = 0;

    sprite_show(0, HARRY_FRONT_A, 2, harry_x, harry_y, 0, 0);
    sprite_show(1, HARRY_FRONT_B, 2, harry_x, harry_y + 16, 0, 0);

    for (;;) {
        flip();

        read_gamepad();

        switch(play_state) {
            case 0:
                if (player1_left()) {
                    if (harry_walk_state == 0) {
                        harry_walk_state = 1;
                        harry_walk_direction = 1;
                        harry_walk_delay = 1;
                    }
                }
                else if (player1_right()) {
                    if (harry_walk_state == 0) {
                        harry_walk_state = 1;
                        harry_walk_direction = 0;
                        harry_walk_delay = 1;
                    }
                }
                else {
                    harry_walk_state = 0;
                }
                break;
        }

        if (harry_walk_delay != 0 && --harry_walk_delay == 0)
            animate_harry();
    }

    return 0;
}
