/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Marco Maccaferri - adapted for game engine usage
 */

#include "graphics_renderer.h"
#include "vga_driver.h"

#include "types.h"

#define RENDER_VRES     0x0100
#define RENDER_BLNK     0x0128
#define RENDER_BASE     0x0129
#define RENDER_OFFSET   0x012A

extern uint32_t binary_graphics_renderer_dat_start[];

RENDER_PARAMS render_param;
RENDER_PARAMS offscreen_param;

TILEMAP tile_map_1[34 * (VRES / 8) + 2];
TILEMAP tile_map_2[34 * (VRES / 8) + 2];

TILEMAP * tile_map;

void start_renderers(uint32_t * palette_def, uint32_t * tiles_def, uint32_t * sprites_def) {
    int i;
    COGLONG * driver = (COGLONG *) binary_graphics_renderer_dat_start;

    driver[RENDER_VRES].src = VRES;

    memset(&tile_map_1, 0x00, sizeof(tile_map_1));

    render_param.tile_map = (uint32_t) &tile_map_1;
    render_param.tiles_def = (uint32_t) tiles_def;
    render_param.sprites_def = (uint32_t) sprites_def;
    render_param.palette_def = (uint32_t) palette_def;
    render_param.x = 0;
    render_param.y = 0;
    memset(&render_param.sprites, 0xFF, sizeof(render_param.sprites));

    offscreen_param.tile_map = (uint32_t) &tile_map_2;
    memcpy(&offscreen_param, &render_param, sizeof(render_param));

    tile_map = (TILEMAP *) &tile_map_2;

    driver[RENDER_BLNK].value = (uint32_t) &line_indicator;
    driver[RENDER_BASE].value = (uint32_t) &scanline_buffer;
    for (i = 0; i < 4; i++) {
        driver[RENDER_OFFSET].value = i;
        cognew(driver, &render_param);
        waitcnt(CNT + 10000);
    }
}

void flip() {
    while (line_indicator == VRES)
        ;
    while (line_indicator != VRES)
        ;

    memcpy(&render_param, &offscreen_param, sizeof(render_param));

    if (tile_map == (TILEMAP *) &tile_map_1) {
        memcpy(&tile_map_2, &tile_map_1, sizeof(tile_map_1));
        offscreen_param.tile_map = (uint32_t) &tile_map_2;
        tile_map = (TILEMAP *) &tile_map_2;
    }
    else {
        memcpy(&tile_map_1, &tile_map_2, sizeof(tile_map_2));
        offscreen_param.tile_map = (uint32_t) &tile_map_1;
        tile_map = (TILEMAP *) &tile_map_1;
    }
}

void sprite_show(uint8_t idx, uint8_t tile, uint8_t palette, uint16_t x, uint16_t y, uint8_t flip, uint8_t mirror) {
    offscreen_param.sprites[idx].tile = tile;
    offscreen_param.sprites[idx].palette = palette;
    offscreen_param.sprites[idx].x = x;
    offscreen_param.sprites[idx].y = y;
    offscreen_param.sprites[idx].flip = flip;
    offscreen_param.sprites[idx].mirror = mirror;
}

void sprite_hide(uint8_t idx) {
    offscreen_param.sprites[idx].value = 0xFFFFFFFF;
}

void sprite_move(uint8_t idx, uint16_t x, uint16_t y) {
    offscreen_param.sprites[idx].x = x;
    offscreen_param.sprites[idx].y = y;
}
