/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Marco Maccaferri - adapted for game engine usage
 */

#ifndef GRAPHICS_RENDERER_H_
#define GRAPHICS_RENDERER_H_

#include <propeller.h>

typedef union {
    uint16_t value;
    struct {
        uint8_t tile;
        uint8_t palette;
    };
} TILEMAP;

typedef union {
    uint32_t value;
    struct {
        uint32_t x       :9;
        uint32_t y       :9;
        uint32_t palette :6;
        uint32_t tile    :6;
        uint32_t mirror  :1;
        uint32_t flip    :1;
    };
} SPRITE_STATE;

typedef struct {
    uint32_t     tile_map;
    uint32_t     tiles_def;
    uint32_t     sprites_def;
    uint32_t     palette_def;
    uint32_t     x;
    uint32_t     y      :9;
    uint32_t     y_lock :9;
    uint32_t     res1   :14;
    SPRITE_STATE sprites[64];
} RENDER_PARAMS;

extern RENDER_PARAMS render_param;
extern RENDER_PARAMS offscreen_param;

extern TILEMAP * tile_map;

void start_renderers(uint32_t * palette_def, uint32_t * tiles_def, uint32_t * sprites_def);
void flip();

void sprite_show(uint8_t idx, uint8_t tile, uint8_t palette, uint16_t x, uint16_t y, uint8_t flip, uint8_t mirror);
void sprite_hide(uint8_t idx);
void sprite_move(uint8_t idx, uint16_t x, uint16_t y);

#endif /* GRAPHICS_RENDERER_H_ */
