/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * SN76489 emulator
 * Based on SNEcog by Johannes Ahlebrand
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Johannes Ahlebrand - Initial implementation
 *     Marco Maccaferri   - Converted to C and adapted for game engine
 */

#include <propeller.h>

#include "types.h"
#include "sn76489.h"

#define ARG1                0x0068
#define ARG2                0x0069
#define R1                  0x006A
#define SAMPLE_RATE         0x006D

extern uint32_t binary_sn76489_dat_start[];

static SN76489_REGISTERS registers;

SN76489_REGISTERS * sn76489_start(int32_t Right, int32_t Left) {
    int32_t cog;
    COGLONG * driver = (COGLONG *) binary_sn76489_dat_start;

    memset(&registers, 0xFF, sizeof(registers));

    driver[ARG1].value = 0x18000000 | Left;
    driver[ARG2].value = 0x18000000 | Right;
    driver[R1].value = ((1 << Right) | (1 << Left)) & 0xFFFFFFFE;
    driver[SAMPLE_RATE].value = CLKFREQ / 176400;

    cog = cognew(driver, &registers);

    return cog != -1 ? &registers : NULL;
}

void sn76489_reset_registers() {
    memset(&registers, 0xFF, sizeof(registers));
}

void sn76489_set_freq(int32_t Channel, int32_t Frequency) {
    registers.ch[Channel].freq = Frequency;
}

void sn76489_set_volume(int32_t Channel, int32_t Volumelevel) {
    registers.ch[Channel].volume = Volumelevel;
}

void sn76489_play(int32_t Channel, int32_t Frequency, int32_t Volumelevel) {
    registers.ch[Channel].freq = Frequency;
    registers.ch[Channel].volume = Volumelevel;
}
