/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * SN76489 emulator
 * Based on SNEcog by Johannes Ahlebrand
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Johannes Ahlebrand - Initial implementation
 *     Marco Maccaferri   - Converted to C and adapted for game engine
 */

#ifndef SN76489_H_
#define SN76489_H_

typedef union {
    uint16_t reg[8];
    struct {
        struct {
            uint16_t freq;
            uint16_t volume;
        } ch[4];
    };
} SN76489_REGISTERS;

/**
 * Starts SNEcog in a single cog
 *
 * Returns a pointer to the first SN register in hub memory
 * on success; otherwise returns 0
 *
 * right - The pin to output the right channel to. 0 = Not used
 * left - The pin to output the left channel to. 0 = Not used
 */
SN76489_REGISTERS * sn76489_start(int32_t Right, int32_t Left);

/**
 * Reset all 8 SN registers
 */
void sn76489_reset_registers();

/**
 * Sets the frequency of a SN channel
 *
 * channel - The SN channel to set (0 - 3)
 * frequency - The 10bit frequency value (0 - 1023)
 */
void sn76489_set_freq(int32_t Channel, int32_t Frequency);

/**
 * Sets the frequency of a SN channel
 *
 * channel - The SN channel to set (0 - 3)
 * volumeLevel - A value betwen 0 and 15 (lo value = hi volume)
 */
void sn76489_set_volume(int32_t Channel, int32_t Volumelevel);

/**
 * Sets the attributes of an SN channel
 *
 * channel - The SN channel to set (0 - 3)
 * frequency - The 10bit frequency value (0 - 1023)
 * volumeLevel - A value betwen 0 and 15 (lo value = hi volume)
 */
void sn76489_play(int32_t Channel, int32_t Frequency, int32_t Volumelevel);

#endif /* SN76489_H_ */
